from __future__ import annotations  # Added for type hints

import struct

from collections import defaultdict
from pathlib import Path
from typing import Any, DefaultDict, Optional, Union

from fittie.fitfile.fitfile import FitFile
from fittie.fitfile.profile.fit_types import FIT_TYPES
from fittie.fitfile.utils.datastream import DataStream, Streamable
from fittie.fitfile.utils.exceptions import DecodeException
from fittie.fitfile.data_message import DataMessage
from fittie.fitfile.definition_message import DefinitionMessage
from fittie.fitfile.field_description import FieldDescription
from fittie.fitfile.header import decode_header
from fittie.fitfile.profile.mesg_nums import MESG_NUMS
from fittie.fitfile.records import read_record_header, read_message


def decode(
    source: Union[str, Path, Streamable], calculate_crc: Optional[bool] = True
) -> FitFile:
    """
    Decode a fit file
    """
    with DataStream(source) as data:
        if not calculate_crc:
            # Don't calculate checksum
            data.should_calculate_crc = False

        header = decode_header(data)

        local_message_definitions: dict[int, DefinitionMessage] = {}
        developer_data: dict[int, dict[str, Any]] = {}
        messages: DefaultDict[str, list[DataMessage]] = defaultdict(list)

        while data.tell() < header.length + header.data_size:
            # Read message
            message = read_message(
                local_message_definitions,
                developer_data,
                data,
            )

            if isinstance(message, DefinitionMessage):
                local_message_definitions[message.header.local_message_type] = message
            else:
                if (
                    global_message_type := local_message_definitions.get(
                        message.header.local_message_type
                    ).global_message_type
                ) == 207:
                    # Add developer data index
                    index = message.fields["developer_data_index"]
                    developer_data[index] = message.fields
                    developer_data[index].update({"fields": {}})
                elif global_message_type == 206:
                    # Add field descriptions
                    index = message.fields["developer_data_index"]
                    field = FieldDescription(**message.fields)
                    developer_data[index]["fields"][
                        field.field_definition_number
                    ] = field

                messages[MESG_NUMS[global_message_type]].append(message)

        calculated_crc = data.calculated_crc
        (crc,) = struct.unpack("H", data.read(2))

        if calculate_crc and crc != calculated_crc:
            raise DecodeException(
                detail=(
                    "the calculated crc does not match the crc at the end of the file"
                ),
                position=data.tell(),
            )

    fitfile = FitFile(
        header=header,
        data_messages=messages,
        local_message_definitions=local_message_definitions,
        developer_data=developer_data,
    )

    return fitfile


def decode_file_type(source: Union[str, Path, Streamable]) -> str:
    """
    Only reads the File header, the first definition message and the first data message
    to retrieve the file type (e.g. activity, workout, weight etc.).

    This only works if the FIT file is encoded according to the FIT protocol best
    practices:
    - FIT file should start with a file header
    - A definition message for 'file_id'
    - A data message for 'file_id'
    """
    fields = {}

    with DataStream(source) as data:
        decode_header(data)
        local_message_definitions = {}
        developer_data = {}

        file_id_definition_message_record_header = read_record_header(data)

        file_id_definition_message: DefinitionMessage = read_message(
            file_id_definition_message_record_header,
            local_message_definitions.get(
                file_id_definition_message_record_header.local_message_type
            ),
            developer_data,
            data,
        )

        if file_id_definition_message.global_message_type != 0:
            raise DecodeException(
                detail=(
                    "provided FIT file is possible not encoded correctly, the first "
                    "definition message does not have global_message_type 0 (file_id)"
                ),
                position=data.tell(),
            )

        local_message_definitions[
            file_id_definition_message_record_header.local_message_type
        ] = file_id_definition_message

        file_id_data_message_record_header = read_record_header(data)
        file_id_data_message: DataMessage = read_message(
            file_id_data_message_record_header,
            local_message_definitions.get(
                file_id_data_message_record_header.local_message_type
            ),
            developer_data,
            data,
        )
        fields = file_id_data_message.fields

    file_type_number = fields["type"]

    if not (file_type := FIT_TYPES["file"]["values"].get(file_type_number)):
        raise DecodeException(
            detail=f"Invalid file type number received: {file_type_number}",
            position=data.tell(),
        )

    return file_type["value_name"]



from __future__ import annotations  # Added for type hints


def get_gear_change_data(data: int) -> dict[str, int]:
    """
    Reads rear gear number, rear gear, front gear number, front gear
    from the provided data.

    No check will be done to verify if the provided data is correct

    Each gear information is 8 bits long
    """
    gear_change_data = {
        'rear_gear_number': data & 255,
        'rear_gear': data >> 8 & 255,
        'front_gear_number': data >> 16 & 255,
        'front_gear': data >> 24 & 255,
    }

    return gear_change_data



from __future__ import annotations  # Added for type hints

import functools
import itertools

from abc import ABC, abstractmethod
from typing import Any, Optional, Iterable

from fittie.fitfile.profile.messages import MESSAGES
from fittie.fitfile.data_message import DataMessage
from fittie.fitfile.definition_message import DefinitionMessage
from fittie.fitfile.header import Header
from fittie.fitfile.profile.fit_types import FIT_TYPES
from fittie.fitfile.profile.mesg_nums import MESG_NUMS
from fittie.fitfile.util import datetime_from_timestamp


class _IterableMixin(ABC):
    @property
    @abstractmethod
    def _iter_collection(self):
        """
        Defines which collection of data the __next__ method should use.

        Make sure to decorate this with @cached_property if it is an expensive
        calculation.

        Also make sure to add __dict__ to  __slots__ if slots are used and the
        property is cached, because  __dict__ has to be mutable.
        """
        ...

    def __next__(self) -> dict:
        if self._iter_index >= len(self._iter_collection):
            # Remove _iter_collection value from cache, if it exists.
            # If _iter_collection is cached, then it should exist and be cleared
            # as soon as the collection has been fully iterated.
            if "_iter_collection" in self.__dict__:
                del self.__dict__["_iter_collection"]

            # Remove _iter_file from dict, because it is only used in iterations
            if "_iter_filter" in self.__dict__:
                del self.__dict__["_iter_filter"]

            raise StopIteration

        value = self._iter_collection[self._iter_index]
        self._iter_index += 1

        if hasattr(self, "_iter_filter") and (fields := self._iter_filter["fields"]):
            return {field: value.fields.get(field) for field in fields}

        return value.fields

    def __iter__(self):
        self._iter_index = 0
        return self


class FitFile(_IterableMixin):

    header: Header
    data_messages: dict[str, list[DataMessage]]
    local_message_definitions: dict[int, DefinitionMessage] = {}
    developer_data: dict[int, dict[str, Any]] = {}  # TODO: add typing

    def __init__(
        self,
        header: Header,
        data_messages: dict[str, list[DataMessage]],
        local_message_definitions: dict[int, DefinitionMessage],
        developer_data: dict[int, dict[str, Any]],
    ):
        self.header = header
        self.data_messages = data_messages
        self.local_message_definitions = local_message_definitions
        self.developer_data = developer_data

    @functools.cached_property
    def _iter_collection(self) -> Iterable[DataMessage]:
        if hasattr(self, "_iter_filter"):
            # Entered through __call__, filter messages
            message_type = self._iter_filter["message_type"]
            return self.data_messages.get(message_type, [])

        return list(itertools.chain(*self.data_messages.values()))

    def __call__(self, *, message_type: str, fields: Optional[list[str]] = None):
        """
        Makes the instance callable and adds filter options for a specific message
        type and keywords of field values to return while iterating
        """
        self._iter_filter = {
            "fields": fields or [],
            "message_type": message_type,
        }

        return self

    @property
    def file_id(self) -> Optional[dict[str, Any]]:
        """
        Get file id information.

        Raw values from the FIT file will be filled with information from the Garmin
        FIT SDK Fit Types
        """
        if not (file_id_messages := self.data_messages.get("file_id")):
            raise ValueError(
                "no file_id message detected, FIT file is possible incorrect"
            )

        file_id = {}

        # Should be just one file_id, but to be sure use latest from list
        for key, value in file_id_messages[-1].fields.items():
            if key == "time_created":
                file_id[key] = datetime_from_timestamp(value)
            elif key == "type":
                file_id[key] = FIT_TYPES["file"]["values"][value]["value_name"]
            elif key == "manufacturer":
                file_id[key] = FIT_TYPES["manufacturer"]["values"][value]["value_name"]
            else:
                file_id[key] = value

        return file_id

    @property
    def file_type(self) -> str:
        """
        Returns the file type of the FIT File. The file type is retrieved from the
        file_id message, which is always present in a FIT file.
        """
        return self.file_id["type"]

    @property
    def available_message_types(self) -> list[str]:
        """Returns a list of all message types that this FIT file contains"""
        return list(self.data_messages.keys())

    @functools.cached_property
    def available_fields(self) -> dict[str, str]:
        """Returns a list of all field names as key, with units as value"""
        fields = {}

        for definition_message in self.local_message_definitions.values():
            _fields = MESSAGES[definition_message.global_message_type]["fields"]

            for field_definition in definition_message.field_definitions:
                field = _fields[field_definition.number]
                fields[field["field_name"]] = field["units"]

        return fields

    def get_messages_by_type(self, message_type: str) -> list[DataMessage]:
        """
        Returns all messages of the provided type, if the FIT file contains these
        messages. If not, it will return an empty list.

        If the provided message type is unknown, a ValueError will be raised.
        """
        if message_type not in MESG_NUMS.values():
            raise ValueError(f"unknown message type '{message_type}' received")

        return self.data_messages.get(message_type, [])



from unittest import mock
from functools import cached_property

from fittie.fitfile.fitfile import _IterableMixin  # noqa


class Message:
    fields = {"a": 1, "b": 2, "c": 3}


def compute() -> list:
    return [Message(), Message(), Message()]


class FooCached(_IterableMixin):
    @cached_property
    def _iter_collection(self):
        return compute()


def test_iterable():
    foo = FooCached()
    lst = [Message(), Message(), Message()]

    with mock.patch(
        "tests.fittie.fitfile.test_iterable.compute", return_value=lst
    ) as mock_compute:
        for value in foo:
            # Assert fields property is returned
            assert value == {"a": 1, "b": 2, "c": 3}

        # Assert property is cached
        mock_compute.assert_called_once()


def test_iterable_fitfile(small_fitfile):
    messages = []

    for message in small_fitfile:
        assert isinstance(message, dict)
        messages.append(message)

    assert len(messages) == 2


def test_iterable_fitfile_filtered_messages(small_fitfile):
    messages = []

    for message in small_fitfile(message_type="file_id"):
        assert isinstance(message, dict)
        messages.append(message)

    assert len(messages) == 1
    assert messages[0] == {
        "manufacturer": 255,
        "number": 0,
        "product": 0,
        "serial_number": None,
        "time_created": 1046114779,
        "type": 4,
    }


def test_iterable_fitfile_filtered_messages_and_fields(small_fitfile):
    messages = []

    for message in small_fitfile(
        message_type="record", fields=["timestamp", "speed", "power", "heart_rate"]
    ):
        assert isinstance(message, dict)
        messages.append(message)

    assert len(messages) == 1
    assert messages[0] == {
        "heart_rate": 98,
        "power": 121,
        "speed": 3765,
        "timestamp": 1046114799,
    }



from typing import Any

from fittie.fitfile.profile.fit_types import FIT_TYPES
from fittie.fitfile.util import datetime_from_timestamp


def enrich_data(fields: dict[str, Any]) -> None:
    """Modifies the field values in place with a lookup value from the FIT sdk"""
    for key, value in fields.items():
        if key == "timestamp":
            fields[key] = datetime_from_timestamp(value)
            continue

        if key in FIT_TYPES:
            try:
                fields[key] = FIT_TYPES[key]["values"][value]["value_name"]
            except KeyError:
                ...



from fittie.fitfile.profile import MessageProfile, FieldProfile, SubField
from fittie.fitfile.profile.messages import MESSAGES
from fittie.fitfile.profile.util import (
    get_message_profile,
    dict_to_message_profile,
    dict_to_field_profile,
)


def test_get_message_profile():
    # Test if all messages can be converted to a MessageProfile
    for number, value in MESSAGES.items():
        assert isinstance(
            message_profile := get_message_profile(number), MessageProfile
        )

        # Test if all fields are of type FieldProfile
        for field_number, field in message_profile.fields.items():
            assert isinstance(field, FieldProfile)

            # Test if all subfields are of type SubField
            for subfield in field.subfields:
                assert isinstance(subfield, SubField)


def test_dict_to_message_profile():
    # Messages number 0 is file_id, this message has fields with subfields
    message = MESSAGES[0]
    message_profile = dict_to_message_profile(MESSAGES[0])

    assert isinstance(message_profile, MessageProfile)
    assert message_profile.name == message["name"]

    # Check fields
    for field in message_profile.fields.values():
        assert isinstance(field, FieldProfile)

    assert message_profile.group == message["group"]


def test_dict_to_field_profile():
    # Field 2 of message 0 has subfields
    field = MESSAGES[0]["fields"][2]
    field_profile = dict_to_field_profile(field)

    assert isinstance(field_profile, FieldProfile)
    assert field_profile.field_name == "product"

    assert field_profile.field_name == field["field_name"]
    assert field_profile.field_type == field["field_type"]
    assert field_profile.components == field["components"]
    assert field_profile.scale == field["scale"]
    assert field_profile.offset == field["offset"]
    assert field_profile.units == field["units"]
    assert field_profile.bits == field["bits"]
    assert field_profile.accumulate == field["accumulate"]

    assert len(field_profile.subfields) == len(field["subfields"])

    for subfield in field_profile.subfields:
        assert isinstance(subfield, SubField)

    assert field_profile.comment == field["comment"]
